#!/usr/bin/env python

import sys
sys.path.append('.')
sys.path.append('..')

import numpy as np
from extra_data import RunDirectory, by_id
from pes_to_spec.model import Model, matching_ids

from itertools import product

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid.inset_locator import (inset_axes, InsetPosition,
                                                  mark_inset)

from typing import Dict, Optional

from time import time_ns
import pandas as pd

SMALL_SIZE = 12
MEDIUM_SIZE = 18
BIGGER_SIZE = 22

plt.rc('font', size=BIGGER_SIZE)         # controls default text sizes
plt.rc('axes', titlesize=BIGGER_SIZE)    # fontsize of the axes title
plt.rc('axes', labelsize=BIGGER_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('ytick', labelsize=BIGGER_SIZE)   # fontsize of the tick labels
plt.rc('legend', fontsize=MEDIUM_SIZE)   # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

def plot_pes(filename: str, pes_raw_int: np.ndarray, first: int, last: int):
    """
    Plot low-resolution spectrum.

    Args:
      filename: Output file name.
      pes_raw_int: Low-resolution spectrum.

    """
    fig = plt.figure(figsize=(16, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(np.arange(first, last), pes_raw_int, c='b', lw=3, label="Low-resolution measurement")
    ax.legend()
    ax.set(title=f"",
           xlabel="ToF index",
           ylabel="Intensity")
    fig.savefig(filename)
    plt.close(fig)

def plot_result(filename: str, spec_pred: Dict[str, np.ndarray], spec_smooth: np.ndarray, spec_raw_pe: np.ndarray, spec_raw_int: Optional[np.ndarray]=None, pes: Optional[np.ndarray]=None, pes_to_show: Optional[str]="", pes_bin: Optional[np.ndarray]=None):
    """
    Plot result with uncertainty band.

    Args:
      filename: Output file name.
      spec_pred: Predicted result with uncertainty bands in a dictionary.
      spec_smooth: Smoothened expected result with shape (features,).
      spec_raw_pe: x axis with the photon energy in eV.
      spec_raw_int: Original true expected result with shape (features,).
      pes: PES spectrum for the inset.
      pes_to_show: Name of the channel shown.
      pes_bin: PES bins.

    """
    fig = plt.figure(figsize=(12, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    unc_stat = np.mean(spec_pred["unc"])
    unc_pca = np.mean(spec_pred["pca"])
    unc = np.sqrt(unc_stat**2 + unc_pca**2)
    ax.plot(spec_raw_pe, spec_smooth, c='b', lw=3, label="High-res. measurement (smoothened)")
    ax.plot(spec_raw_pe, spec_pred["expected"], c='r', ls='--', lw=3, label="High-res. prediction")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc, spec_pred["expected"] + unc, facecolor='green', alpha=0.6, label="68% unc.")
    ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc, spec_pred["expected"] + unc, facecolor='gold', alpha=0.5, label="68% unc.")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc_stat, spec_pred["expected"] + unc_stat, facecolor='red', alpha=0.6, label="68% unc. (stat.)")
    #ax.fill_between(spec_raw_pe, spec_pred["expected"] - unc_pca, spec_pred["expected"] + unc_pca, facecolor='magenta', alpha=0.6, label="68% unc. (syst., PCA)")
    #if spec_raw_int is not None:
    #    ax.plot(spec_raw_pe, spec_raw_int, c='b', lw=1, ls='--', label="High-resolution measurement")
    Y = np.amax(spec_smooth)
    ax.legend(frameon=False, borderaxespad=0, loc='upper left')
    ax.set(title=f"", #avg(stat unc) = {unc_stat}, avg(pca unc) = {unc_pca}",
           xlabel="Photon energy [eV]",
           ylabel="Intensity",
           ylim=(0, 1.2*Y))
    if pes is not None:
        ax2 = plt.axes([0,0,1,1])
        # Manually set the position and relative size of the inset axes within ax1
        ip = InsetPosition(ax, [0.65,0.6,0.35,0.4])
        ax2.set_axes_locator(ip)
        Ypes = np.amax(pes)
        ax2.plot(pes_bin, pes, c='black', lw=3)
        ax2.set(title=f"Low-resolution example data",
                xlabel="Bin",
                ylabel=f"{pes_to_show}",
                ylim=(0, 1.2*Ypes),
                #labelsize=SMALL_SIZE,
                #xticklabels=dict(fontdict=dict(fontsize=SMALL_SIZE)),
                #yticklabels=dict(fontdict=dict(fontsize=SMALL_SIZE)),
                )
        ax2.title.set_size(SMALL_SIZE)
        ax2.xaxis.label.set_size(SMALL_SIZE)
        ax2.yaxis.label.set_size(SMALL_SIZE)
        ax2.tick_params(axis='both', which='major', labelsize=SMALL_SIZE)
    fig.savefig(filename)
    plt.close(fig)

def main():
    """
    Main entry point. Reads some data, trains and predicts.
    """
    run_dir = "/gpfs/exfel/exp/SA3/202121/p002935/raw"
    run_id = "r0015"
    # get run
    run = RunDirectory(f"{run_dir}/{run_id}")

    # get train IDs and match them, so we are sure to have information from all needed sources
    # in this example, there is an offset of -2 in the SPEC train ID, so correct for it
    spec_offset = -2
    spec_tid = spec_offset + run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                                 "data.trainId"].ndarray()
    pes_tid = run['SA3_XTD10_PES/ADC/1:network',
                  "digitizers.trainId"].ndarray()
    xgm_tid = run['SA3_XTD10_XGM/XGM/DOOCS:output',
                  "data.trainId"].ndarray()
    # these are the train ID intersection
    # this could have been done by a select call in the RunDirectory, but it would not correct for the spec_offset
    tids = matching_ids(spec_tid, pes_tid, xgm_tid)
    train_tids = tids[:-10]
    test_tids = tids[-10:]

    # read the spec photon energy and intensity
    spec_raw_pe = run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                      "data.photonEnergy"].select_trains(by_id[tids - spec_offset]).ndarray()
    spec_raw_int = run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                       "data.intensityDistribution"].select_trains(by_id[tids - spec_offset]).ndarray()

    # read the PES data for each channel
    channels = [f"channel_{i}_{l}"
                for i, l in product(range(1, 5), ["A", "B", "C", "D"])]
    pes_raw = {ch: run['SA3_XTD10_PES/ADC/1:network',
                   f"digitizers.{ch}.raw.samples"].select_trains(by_id[tids]).ndarray()
               for ch in channels}
    pes_raw_t = {ch: run['SA3_XTD10_PES/ADC/1:network',
                   f"digitizers.{ch}.raw.samples"].select_trains(by_id[test_tids]).ndarray()
               for ch in channels}

    # read the XGM information
    #xgm_pressure = run['SA3_XTD10_XGM/XGM/DOOCS', "pressure.pressureFiltered.value"].select_trains(by_id[tids]).ndarray()
    #xgm_pe =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()
    #retvol_raw = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[tids]).ndarray()
    #retvol_raw_timestamp = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.timestamp"].select_trains(by_id[tids]).ndarray()

    t = list()
    t_names = list()

    # these have been manually selected:
    #useful_channels = ["channel_1_D",
    #                  "channel_2_B",
    #                  "channel_3_A",
    #                  "channel_3_B",
    #                  "channel_4_C",
    #                  "channel_4_D"]
    model = Model()

    train_idx = np.isin(tids, train_tids)

    model.debug_peak_finding(pes_raw, "test_peak_finding.png")
    print("Fitting")
    start = time_ns()
    model.fit({k: v[train_idx, :]
               for k, v in pes_raw.items()},
              spec_raw_int[train_idx, :],
              spec_raw_pe[train_idx, :])
    t += [time_ns() - start]
    t_names += ["Fit"]

    print("Saving the model")
    start = time_ns()
    model.save("model.joblib")
    t += [time_ns() - start]
    t_names += ["Save"]

    print("Loading the model")
    start = time_ns()
    model = Model.load("model.joblib")
    t += [time_ns() - start]
    t_names += ["Load"]

    print("Check consistency")
    start = time_ns()
    rmse = model.check_compatibility(pes_raw_t)
    print("Consistency check RMSE ratios:", rmse)
    t += [time_ns() - start]
    t_names += ["Consistency"]

    # test
    print("Predict")
    start = time_ns()
    spec_pred = model.predict(pes_raw)
    t += [time_ns() - start]
    t_names += ["Predict"]

    print("Time taken in ms")
    df_time = pd.DataFrame(data=dict(time=t, name=t_names))
    df_time.time *= 1e-6
    print(df_time)

    print("Plotting")
    spec_smooth = model.preprocess_high_res(spec_raw_int)
    first, last = model.get_low_resolution_range()
    first += 10
    last -= 100
    pes_to_show = 'channel_1_D'
    # plot
    for tid in test_tids:
        idx = np.where(tid==tids)[0][0]
        plot_result(f"test_{tid}.png",
                   {k: item[idx, ...] if k != "pca"
                       else item[0, ...]
                       for k, item in spec_pred.items()},
                    spec_smooth[idx, :],
                    spec_raw_pe[idx, :],
                    spec_raw_int[idx, :],
                    pes=-pes_raw[pes_to_show][idx, first:last],
                    pes_to_show=pes_to_show.replace('_', ' '),
                    pes_bin=np.arange(first, last)
                    )
        for ch in channels:
            plot_pes(f"test_pes_{tid}_{ch}.png", pes_raw[ch][idx, first:last], first, last)

if __name__ == '__main__':
    main()
